@@ -4,6 +4,7 @@ import asyncio
 import io
 import os
 import sys
+from logging import Logger
 
 import shellous
 import shellous.redirect as redir
@@ -12,7 +13,7 @@ from shellous.harvest import harvest, harvest_results
 from shellous.log import LOG_DETAIL, LOG_ENTER, LOG_EXIT, LOGGER, log_method, log_timer
 from shellous.redirect import Redirect
 from shellous.result import Result, make_result
-from shellous.util import close_fds, verify_dev_fd
+from shellous.util import close_fds, verify_dev_fd, wait_pid
 
 _KILL_TIMEOUT = 3.0
 _CLOSE_TIMEOUT = 0.25
@@ -21,6 +22,8 @@ _KILL_EXIT_CODE = -9 if sys.platform != "win32" else 1
 _FLAKY_EXIT_CODE = 255
 _BSD = sys.platform.startswith("freebsd") or sys.platform == "darwin"
 
+AUDIT_EVENT_SUBPROCESS_SPAWN = "byllyfish/shellous.subprocess_spawn"
+
 
 def _is_cancelled(ex):
     return isinstance(ex, asyncio.CancelledError)
@@ -38,7 +41,7 @@ def _is_write_mode(cmd):
     return cmd.options.write_mode
 
 
-class _RunOptions:
+class _RunOptions:  # pylint: disable=too-many-instance-attributes
     """_RunOptions is context manager to assist in running a command.
 
     This class sets up low-level I/O redirection and helps close open file
@@ -79,9 +82,11 @@ class _RunOptions:
         "Make sure those file descriptors are cleaned up."
         self.close_fds()
         if exc_value:
-            LOGGER.info(
+            LOGGER.warning(
                 "_RunOptions.exit %r exc_value=%r", self.command.name, exc_value
             )
+            for subcmd in self.subcmds:
+                _cleanup(subcmd)
 
     def _setup_proc_sub(self):
         "Set up process substitution."
@@ -125,18 +130,20 @@ class _RunOptions:
         options = self.command.options
 
         stdin, input_bytes = self._setup_input(
-            options.input, options.input_close, self.encoding
+            Redirect.from_default(options.input, 0, options.pty),
+            options.input_close,
+            self.encoding,
         )
 
         stdout = self._setup_output(
-            options.output,
+            Redirect.from_default(options.output, 1, options.pty),
             options.output_append,
             options.output_close,
             sys.stdout,
         )
 
         stderr = self._setup_output(
-            options.error,
+            Redirect.from_default(options.error, 2, options.pty),
             options.error_append,
             options.error_close,
             sys.stderr,
@@ -181,12 +188,14 @@ class _RunOptions:
         if isinstance(input_, (bytes, bytearray)):
             input_bytes = input_
         elif isinstance(input_, os.PathLike):
-            stdin = open(input_, "rb")
+            stdin = open(input_, "rb")  # pylint: disable=consider-using-with
             self.open_fds.append(stdin)
         elif isinstance(input_, Redirect) and input_.is_custom():
             # Custom support for Redirect constants.
             if input_ == Redirect.INHERIT:
                 stdin = sys.stdin
+            elif input_ == Redirect.IGNORE:
+                input_bytes = None
             else:
                 # CAPTURE uses stdin == PIPE.
                 assert input_ == Redirect.CAPTURE
@@ -198,7 +207,7 @@ class _RunOptions:
         else:
             if encoding is None:
                 raise TypeError("when encoding is None, input must be bytes")
-            input_bytes = input_.encode(encoding)
+            input_bytes = input_.encode(*encoding.split(maxsplit=1))
 
         return stdin, input_bytes
 
@@ -210,8 +219,7 @@ class _RunOptions:
 
         if isinstance(output, (str, bytes, os.PathLike)):
             mode = "ab" if append else "wb"
-            # FIXME: we really just need the file descriptor...
-            stdout = open(output, mode=mode)
+            stdout = open(output, mode=mode)  # pylint: disable=consider-using-with
             self.open_fds.append(stdout)
         elif isinstance(output, Redirect) and output.is_custom():
             # Custom support for Redirect constants.
@@ -226,7 +234,7 @@ class _RunOptions:
             stdout = output
             if close:
                 self.open_fds.append(stdout)
-        elif isinstance(output, (io.StringIO, io.BytesIO, bytearray)):
+        elif isinstance(output, (io.StringIO, io.BytesIO, bytearray, Logger)):
             pass
         elif isinstance(output, io.IOBase):
             # Client-managed File-like object.
@@ -242,10 +250,18 @@ class _RunOptions:
         Initializes `self.pty_fds`.
         """
 
-        self.pty_fds, child_fd = pty_util.open_pty(pty)
-        self.open_fds.append(child_fd)
+        self.pty_fds = pty_util.open_pty(pty)
+        child_fd = int(self.pty_fds.child_fd)
 
-        LOGGER.info("_setup_pty1: %r child_fd=%r", self.pty_fds, child_fd)
+        # On BSD-derived systems like FreeBSD and Darwin, we delay closing the
+        # pty's child_fd in the parent process until after the first read
+        # succeeds. On Linux, we close the child_fd as soon as possible.
+
+        if not _BSD:
+            self.open_fds.append(self.pty_fds.child_fd)
+
+        if LOG_DETAIL:
+            LOGGER.info("_setup_pty1: %r", self.pty_fds)
 
         if stdin == asyncio.subprocess.PIPE:
             stdin = child_fd
@@ -290,6 +306,13 @@ class Runner:
         "Return the command being run."
         return self.options.command
 
+    @property
+    def pid(self):
+        "Return the command's process ID."
+        if not self.proc:
+            return None
+        return self.proc.pid
+
     def result(self, output_bytes=b""):
         "Check process exit code and raise a ResultError if necessary."
         if self.proc:
@@ -319,6 +342,10 @@ class Runner:
         self.tasks.append(task)
         return task
 
+    def _is_bsd_pty(self):
+        "Return true if we're running a pty on BSD."
+        return _BSD and self.options.pty_fds
+
     @log_method(LOG_DETAIL)
     async def _wait(self):
         "Normal wait for background I/O tasks and process to finish."
@@ -327,8 +354,8 @@ class Runner:
         try:
             if self.tasks:
                 await harvest(*self.tasks, trustee=self)
-            if _BSD and self.options.pty_fds:
-                await self._wait_pid()
+            if self._is_bsd_pty():
+                await self._waiter()
 
         except asyncio.CancelledError:
             LOGGER.info("Runner.wait cancelled %r", self)
@@ -339,14 +366,11 @@ class Runner:
     @log_method(LOG_DETAIL)
     async def _wait_pid(self):
         "Manually poll `waitpid` until process finishes."
-        proc_pid = self.proc.pid
-        assert proc_pid
-
+        assert self._is_bsd_pty()
         while True:
-            pid, status = os.waitpid(proc_pid, os.WNOHANG)
-            LOGGER.info("waitpid returned %r", (pid, status))
-
-            if pid == proc_pid:
+            status = wait_pid(self.proc.pid)
+            if status is not None:
+                # pylint: disable=protected-access
                 self.proc._transport._returncode = status
                 self.proc._transport._proc.returncode = status
                 break
@@ -362,16 +386,15 @@ class Runner:
         cancel_signal = self.command.options.cancel_signal
 
         try:
-            done = self.proc.returncode is not None
-
             # If not already done, send cancel signal.
-            if not done:
+            if self.proc.returncode is None:
                 self._send_signal(cancel_signal)
 
             if self.tasks:
                 await harvest(*self.tasks, timeout=cancel_timeout, trustee=self)
-            elif not done:
-                await harvest(self.proc.wait(), timeout=cancel_timeout, trustee=self)
+
+            if self.proc.returncode is None:
+                await harvest(self._waiter(), timeout=cancel_timeout, trustee=self)
 
         except (asyncio.CancelledError, asyncio.TimeoutError) as ex:
             LOGGER.warning("Runner.kill %r (ex)=%r", self, ex)
@@ -403,8 +426,8 @@ class Runner:
             return
 
         try:
-            self.proc.kill()
-            await harvest(self.proc.wait(), timeout=_KILL_TIMEOUT, trustee=self)
+            self._send_signal(None)
+            await harvest(self._waiter(), timeout=_KILL_TIMEOUT, trustee=self)
         except asyncio.TimeoutError as ex:
             LOGGER.error("%r failed to kill process %r", self, self.proc)
             raise RuntimeError(f"Unable to kill process {self.proc!r}") from ex
@@ -428,7 +451,7 @@ class Runner:
         try:
             # Set up subprocess arguments and launch subprocess.
             with self.options as opts:
-                await self._subprocess_exec(opts)
+                await self._subprocess_spawn(opts)
 
             stdin = self.proc.stdin
             stdout = self.proc.stdout
@@ -473,13 +496,13 @@ class Runner:
         self.stderr = stderr
 
         # Add a task to monitor for when the process finishes.
-        if not (_BSD and opts.pty_fds):
+        if not self._is_bsd_pty():
             self.add_task(self._waiter(), "waiter")
 
         return self
 
     @log_method(LOG_DETAIL)
-    async def _subprocess_exec(self, opts):
+    async def _subprocess_spawn(self, opts):
         "Start the subprocess and assign to `self.proc`."
 
         # Second half of pty setup.
@@ -489,6 +512,8 @@ class Runner:
                 pty_util.patch_child_watcher()
 
         with log_timer("asyncio.create_subprocess_exec"):
+            # AUDIT: subprocess spawn
+            sys.audit(AUDIT_EVENT_SUBPROCESS_SPAWN, opts.args[0])
             self.proc = await asyncio.create_subprocess_exec(
                 *opts.args,
                 **opts.kwd_args,
@@ -501,7 +526,10 @@ class Runner:
     @log_method(LOG_DETAIL)
     async def _waiter(self):
         "Run task that waits for process to exit."
-        await self.proc.wait()
+        if self._is_bsd_pty():
+            await self._wait_pid()
+        else:
+            await self.proc.wait()
 
     def _setup_output_sink(self, stream, sink, encoding, tag):
         "Set up a task to write to custom output sink."
@@ -516,6 +544,9 @@ class Runner:
         elif isinstance(sink, bytearray):
             self.add_task(redir.copy_bytearray(stream, sink), tag)
             stream = None
+        elif isinstance(sink, Logger):
+            self.add_task(redir.copy_logger(stream, sink, encoding), tag)
+            stream = None
         return stream
 
     @log_method(LOG_EXIT, exc_value=2)
@@ -701,7 +732,7 @@ class PipeRunner:  # pylint: disable=too-many-instance-attributes
             LOGGER.warning("PipeRunner enter %r ex=%r", self, ex)
             if _is_cancelled(ex):
                 self.cancelled = True
-            await self._wait(kill=True)  # FIXME
+            await self._wait(kill=True)
             raise
 
     @log_method(LOG_EXIT, exc_value=2)
@@ -841,9 +872,13 @@ class PipeRunner:  # pylint: disable=too-many-instance-attributes
 
 def _is_multiple_capture(cmd):
     "Return true if stdin is CAPTURE or both stdout and stderr are CAPTURE."
-    return cmd.options.input == Redirect.CAPTURE or (
-        cmd.options.output == Redirect.CAPTURE and cmd.options.error == Redirect.CAPTURE
-    )
+    input_ = Redirect.from_default(cmd.options.input, 0, cmd.options.pty)
+    if input_ == Redirect.CAPTURE:
+        return True
+
+    output = Redirect.from_default(cmd.options.output, 1, cmd.options.pty)
+    error = Redirect.from_default(cmd.options.error, 2, cmd.options.pty)
+    return output == Redirect.CAPTURE and error == Redirect.CAPTURE
 
 
 def _cleanup(command):
